
package com.trilead.ssh2;

/**
 * An abstract marker interface implemented by all proxy data implementations.
 * 
 * @see HTTPProxyData
 * 
 * @author Christian Plattner, plattner@trilead.com
 * @version $Id: ProxyData.java,v 1.1 2007/10/15 12:49:56 cplattne Exp $
 */

public abstract interface ProxyData
{
}
